"use strict";

Object.defineProperty(exports, "__esModule", { value: true });
const util_1 = require("./util");
const Address_1 = require("./Address");
const elliptic_1 = require("elliptic");
const secp256k1 = new elliptic_1.ec('secp256k1');
class Message {
    constructor(message) {
        this._hash = null;
        this._message = message;
    }
    get message() {
        return this._message;
    }
    get hash() {
        if (this._hash) {
            return this._hash;
        }
        const prefix = Buffer.from(`\x19Ethereum Signed Message:\n${this._message.length}`, 'utf8');
        const messageToSign = Buffer.concat([prefix, this._message]);
        this._hash = util_1.keccak256(messageToSign);
        return this._hash;
    }
    sign(privateKey) {
        const sig = secp256k1.keyFromPrivate(privateKey).sign(this.hash, { canonical: true });
        const sigBuf = Buffer.alloc(65);
        let offset = 0;
        // copy r
        const r = util_1.bnToBuffer(sig.r);
        offset += 32 - r.length;
        offset += r.copy(sigBuf, offset);
        // copy s
        const s = util_1.bnToBuffer(sig.s);
        offset += 32 - s.length;
        offset += s.copy(sigBuf, offset);
        // copy v
        sigBuf.writeUInt8((sig.recoveryParam || 0) + 27, offset);
        return sigBuf;
    }
    ecRecover(signature) {
        if (signature.length !== 65) {
            throw new Error('Invalid signature');
        }
        const recoveryParam = signature[signature.length - 1] - 27;
        const sig = {
            r: signature.slice(0, 32),
            s: signature.slice(32, 64)
        };
        const point = secp256k1.recoverPubKey(this.hash, sig, recoveryParam);
        const pubKey = Buffer.from(point.encode('hex', true), 'hex');
        return Address_1.Address.from(pubKey).address;
    }
}
exports.Message = Message;