"use strict";

Object.defineProperty(exports, "__esModule", { value: true });
const util_1 = require("./util");
const elliptic_1 = require("elliptic");
const rlp_1 = require("rlp");
const secp256k1 = new elliptic_1.ec('secp256k1');
class Transaction {
    constructor(params) {
        this.nonce = util_1.numberToBuffer(params.nonce);
        this.gasPriceWei = util_1.bigNumberToBuffer(params.gasPriceWei);
        this.gasLimit = util_1.bigNumberToBuffer(params.gasLimit);
        this.toAddress = params.toAddress ? util_1.hexToBuffer(params.toAddress) : Buffer.alloc(0);
        this.valueWei = util_1.bigNumberToBuffer(params.valueWei);
        this.data = params.data ? util_1.hexToBuffer(params.data) : Buffer.alloc(0);
        this.chainId = params.chainId ? params.chainId : undefined; // disallow 0
        this.v = params.v || 28;
        this.r = params.r ? util_1.hexToBuffer(params.r) : Buffer.alloc(0);
        this.s = params.s ? util_1.hexToBuffer(params.s) : Buffer.alloc(0);
    }
    get fields() {
        return [this.nonce, this.gasPriceWei, this.gasLimit, this.toAddress, this.valueWei, this.data, util_1.numberToBuffer(this.v), this.r, this.s // 8: s
        ];
    }
    get rlp() {
        return rlp_1.default.encode(this.fields);
    }
    get hash() {
        return util_1.keccak256(this.rlp);
    }
    get fieldsForSigning() {
        const fields = this.fields.slice(0, 6);
        return this.chainId ? fields.concat([
        // EIP155
        util_1.numberToBuffer(this.chainId), util_1.numberToBuffer(0), util_1.numberToBuffer(0) // 8: s = 0
        ]) : fields;
    }
    get hashForSigning() {
        return util_1.keccak256(rlp_1.default.encode(this.fieldsForSigning));
    }
    sign(privateKey) {
        const sig = secp256k1.keyFromPrivate(privateKey).sign(this.hashForSigning, { canonical: true });
        this.v = (sig.recoveryParam || 0) + 27 + (this.chainId ? this.chainId * 2 + 8 : 0);
        this.r = util_1.bnToBuffer(sig.r);
        this.s = util_1.bnToBuffer(sig.s);
        return this.rlp;
    }
}
exports.Transaction = Transaction;